/*****************************************************************************
  addhardware.c -- Routines to modify the graph by adding DFT hardware
  Define CONVERT to gate gate names put into the circuit graph
*****************************************************************************/
#include "all.h"
#include <string.h>
// Logic circuit constructor

CIRCUIT * new_circuit (char * name)
{ // CIRCUIT
  CIRCUIT * new_ckt;

  new_ckt = (CIRCUIT *) malloc (sizeof (CIRCUIT));
  new_ckt->Pi_list = NULL;
  new_ckt->Gate_list = NULL;
  new_ckt->Po_list = NULL;
  new_ckt->InOut_list = NULL;
  new_ckt->Reg_list = NULL;
  new_ckt->nmbr_Gates = 0;
  new_ckt->nmbr_PIs = 0;
  new_ckt->nmbr_POs = 0;
  new_ckt->nmbr_INOUTs = 0;
  new_ckt->nmbr_REGs= 0;
  new_ckt->negative_gate_nmbr = 0;
  new_ckt->highest_level = 0;
  new_ckt->gate_alloc = 0;
  new_ckt->pi_alloc = 0;
  new_ckt->po_alloc = 0;
  new_ckt->inout_alloc = 0;
  new_ckt->reg_alloc = 0;
  new_ckt->design_name = name;
  new_ckt->highest_level = 0;
  return (new_ckt);
} // CIRCUIT


void free_gate(CIRCUIT *ckt, NODE *gate)
{ /* free_gate */
  
  if (nmbroutputs(gate) > 0)
    {
      free(fanouts(gate));
    }
  if (nmbrinputs(gate) > 0)
    {
    free(fanins(gate));
  }

  free(gate);
} /* free_gate */

void free_circuit (CIRCUIT *ckt)
{ /* free_circuit */
  int i;
  
  for (i=0; i<nmbr_pi(ckt); i++)
    {
      free_gate(ckt, PIs(ckt, i));
    }
  free(pi_list(ckt));
  for (i=0; i<nmbr_gates(ckt); i++)
    {
      free_gate(ckt, gatesx(ckt, i));
    }
  free(gate_list(ckt));
  for (i=0; i<nmbr_po(ckt); i++)
    {
      free_gate(ckt, POs(ckt, i));
    }
  free(po_list(ckt));
  free(ckt);
} /* free_circuit */

void move_gate (CIRCUIT *ckt, int from_index, int to_index)
{ /* move_gate */

  if(from_index == to_index)
    return;
  
  gatesx (ckt, to_index) = gatesx (ckt, from_index);
  gatesx (ckt, from_index) = NULL;
  gate_index (gatesx (ckt, to_index)) = to_index;
  return;
} /* move_gate */

void delete_logic_gate (CIRCUIT *ckt, NODE * gate)
{ /* delete_logic_gate */

  move_gate (ckt, nmbr_gates (ckt) - 1, gate_index (gate));
  nmbr_gates (ckt) --;
  free_gate (ckt, gate);

} /* delete_logic_gate */

void free_pi (CIRCUIT *ckt, NODE *unwanted_pi)
{ /* free_pi */
  
  if (nmbroutputs(unwanted_pi) > 0)
    {
      free(fanouts(unwanted_pi));
    }
  free (unwanted_pi);
  return;
} /* free_pi */

void move_pi (CIRCUIT *ckt, int from_index, int to_index)
{ /* move_pi */

  if(from_index == to_index)
    return;
  
  PIs (ckt, to_index) = PIs (ckt, from_index);
  PIs (ckt, from_index) = NULL;
  gate_index (PIs (ckt, to_index)) = to_index;
  return;
} /* move_pi */

void delete_pi (CIRCUIT *ckt, NODE * gate)
{ /* delete_pi */

  move_pi (ckt, nmbr_pi (ckt) - 1, gate_index (gate));
  nmbr_pi (ckt) --;
  free_pi (ckt, gate);

} /* delete_pi */

void free_po (CIRCUIT *ckt, NODE *unwanted_po)
{ /* free_po */
  
  if (nmbrinputs (unwanted_po) > 0)
    {
      free (fanins (unwanted_po));
    }
  free(unwanted_po);
  return;
} /* free_po */

void move_po (CIRCUIT *ckt, int from_index, int to_index)
{ /* move_po */

  if(from_index == to_index)
    return;
  
  POs (ckt, to_index) = POs (ckt, from_index);
  POs (ckt, from_index) = NULL;
  gate_index (POs (ckt, to_index)) = to_index;
  return;
} /* move_po */

void delete_po (CIRCUIT *ckt, NODE * gate)
{ /* delete_po */

  move_po (ckt, nmbr_po (ckt) - 1, gate_index (gate));
  nmbr_po (ckt) --;
  free_po (ckt, gate);

} /* delete_po */

/* Routine to reallocate Node lists */
NODE **reallocate_nodes(NODE **old_list, int old_size, int new_size)
{ /* reallocate_nodes */
  NODE **new_list;
  int i;
  
  if (new_size < old_size)
    {
      fprintf(resultsout,
	      "reallocate_nodes: Trying to reallocate fewer nodes.\n");
      return (old_list);
    }

  new_list = (NODE **)calloc(new_size, sizeof(NODE *));
  if(new_list == NULL)
    {
      //fatal_error("reallocate_nodes", MEMORY_ALLOC_FAILURE, NULL);
      /* WARNING * -- Removed fatal_error here, need fatalerror.c! */
	printf("reallocate_nodes -- need fatal_error routine!\n");
    }

  for (i=0; i < old_size; i++)
    new_list[i] = old_list[i];
  
  free(old_list);

  return(new_list);
} /* reallocate_nodes */

/***************************************************************************
  make_new_pi() allocates and initializes the necessary variables for
                new TEST PI
***************************************************************************/
NODE * make_new_pi (CIRCUIT *ckt, int gate_no, char * gate_name)
{ /* make_new_pi */
  NODE *gate;

  ++ nmbr_pi (ckt);
  if (nmbr_pi (ckt) > ckt->pi_alloc)
    {
      ckt->pi_alloc = 2 * ckt->pi_alloc;
      if (ckt->pi_alloc == 0) ckt->pi_alloc = 20;
      if (pi_list (ckt) == NULL)
	pi_list (ckt) = (NODE **) calloc (ckt->pi_alloc, sizeof (NODE *));
      else
	pi_list (ckt) = (NODE **) realloc (pi_list (ckt),
					   sizeof (NODE *) * ckt->pi_alloc);
    }
  gate = (NODE *) calloc(1, sizeof(NODE));
  PIs(ckt, nmbr_pi(ckt) - 1) = gate;

#if (defined(VERILOG))
  gate->comp_name = strdup (gate_name);
#endif
  gate_number(gate) = gate_no;
  gatefunction (gate) = p_pi;
  /*  outputindex (gate) = 0; */
  nmbrinputs (gate) = 0;
  nmbroutputs (gate) = 0;
  levelnumber (gate) = 1;
  gate_index (gate) = nmbr_pi (ckt) - 1;
  return (gate);

} /* make_new_pi */

/***************************************************************************
  make_new_logic_gate() allocates and initializes the necessary variables
                       for parityflipper
***************************************************************************/
NODE * make_new_logic_gate (CIRCUIT * ckt,           /* Circuit graph */
			    P_GATETYPE function,
			    int gate_no,
			    char *gate_name,
			    int level)
{ /* make_new_logic_gate */
  NODE *my_new_gate;

  ++nmbr_gates (ckt);
  if (nmbr_gates (ckt) > ckt->gate_alloc)
    {
      ckt->gate_alloc = 2 * ckt->gate_alloc;
      if (ckt->gate_alloc == 0) ckt->gate_alloc = 100;
      if (gate_list (ckt) == NULL)
	gate_list(ckt) =  (NODE **) calloc (ckt->gate_alloc, sizeof (NODE *));
      else
	gate_list(ckt) = (NODE **) realloc (gate_list (ckt),
				      sizeof (NODE *) * ckt->gate_alloc);
    }
  my_new_gate = (NODE *) calloc (1, sizeof (NODE));
  gatesx (ckt, nmbr_gates (ckt) - 1) = my_new_gate;

#if (defined(VERILOG))
  if (gate_name != NULL)
    my_new_gate->comp_name = strdup (gate_name);
  else
    my_new_gate->comp_name = NULL;
#endif
  gatefunction (my_new_gate) = function;
  gate_number (my_new_gate) = gate_no;
  nmbroutputs (my_new_gate) = 0;
  /*  outputindex (my_new_gate) = 0; */
  fanouts (my_new_gate) = NULL;
  nmbrinputs (my_new_gate) = 0;
  fanins (my_new_gate) = NULL;
  levelnumber (my_new_gate) = level;
  gate_index (my_new_gate) = nmbr_gates (ckt) - 1;
  return (my_new_gate);

} /* make_new_logic_gate */
 
/****************************************************************************/
/* add a new primary output */
/****************************************************************************/
NODE * make_new_po (CIRCUIT * ckt,
		    int gate_no,
		    char * gate_name,
		    int level)
{ /* make_new_po */
  int new_po = 0;
  NODE * newpogate;
  NODE ** tmp_gate_list;

  tmp_gate_list = po_list (ckt);
  if (nmbr_po (ckt) + 1 > ckt->po_alloc)
    {
      ckt->po_alloc = 2 * ckt->po_alloc;
      if (ckt->po_alloc == 0) ckt->po_alloc = 20;
      if (po_list (ckt) == NULL)
	po_list (ckt) = (NODE **) calloc (ckt->po_alloc, sizeof (NODE *));
      else
	po_list (ckt) = (NODE **) realloc (tmp_gate_list,
					   sizeof (NODE *) * ckt->po_alloc);
    }
  new_po = nmbr_po (ckt);
  /* Allocate storage for the new PO record */
  POs (ckt, new_po) = (NODE *) calloc (1, sizeof (NODE));
  ++ nmbr_po (ckt);
  newpogate = POs (ckt, new_po);

  gatefunction (newpogate) = p_po;
  gate_number (newpogate) = gate_no;
#if (defined(VERILOG))
  newpogate->comp_name = strdup (gate_name);
#endif
  levelnumber (newpogate) = level;
  nmbroutputs (newpogate) = 0;
  /*  outputindex (newpogate) = 0; */
  gate_index (newpogate) = nmbr_po (ckt) - 1;
  fanouts (newpogate) = NULL;
  nmbrinputs (newpogate) = 0;
  fanins (newpogate) = NULL;
  return (newpogate);

} /* make_new_po */

/***************************************************************************
  make_new_inout() allocates and initializes the necessary variables for
                   new INOUT port
***************************************************************************/
NODE * make_new_inout (CIRCUIT *ckt, int gate_no, char * gate_name)
{ /* make_new_inout */
  NODE *gate;

  ++ nmbr_inout (ckt);
  if (nmbr_inout (ckt) > ckt->inout_alloc)
    {
      ckt->inout_alloc = 2 * ckt->inout_alloc;
      if (ckt->inout_alloc == 0) ckt->inout_alloc = 20;
      if (inout_list (ckt) == NULL)
	inout_list (ckt) = (NODE **) calloc (ckt->inout_alloc, sizeof (NODE *));
      else
	inout_list (ckt) = (NODE **) realloc (inout_list (ckt),
					   sizeof (NODE *) * ckt->inout_alloc);
    }
  gate = (NODE *) calloc(1, sizeof(NODE));
  INOUTs(ckt, nmbr_inout(ckt) - 1) = gate;

#if (defined(VERILOG))
  gate->comp_name = strdup (gate_name);
#endif
  gate_number(gate) = gate_no;
  gatefunction (gate) = p_inout;
  /*  outputindex (gate) = 0; */
  nmbrinputs (gate) = 0;
  nmbroutputs (gate) = 0;
  levelnumber (gate) = 1;
  gate_index (gate) = nmbr_inout (ckt) - 1;
  return (gate);

} /* make_new_inout */

/***************************************************************************
  make_new_reg() allocates and initializes the necessary variables for
                 new REG declaration
***************************************************************************/
NODE * make_new_reg (CIRCUIT *ckt, int gate_no, char * gate_name)
{ /* make_new_reg */
  NODE *gate;

  ++ nmbr_reg (ckt);
  if (nmbr_reg (ckt) > ckt->reg_alloc)
    {
      ckt->reg_alloc = 2 * ckt->reg_alloc;
      if (ckt->reg_alloc == 0) ckt->reg_alloc = 20;
      if (reg_list (ckt) == NULL)
	reg_list (ckt) = (NODE **) calloc (ckt->reg_alloc, sizeof (NODE *));
      else
	reg_list (ckt) = (NODE **) realloc (reg_list (ckt),
					   sizeof (NODE *) * ckt->reg_alloc);
    }
  gate = (NODE *) calloc(1, sizeof(NODE));
  REGs(ckt, nmbr_reg(ckt) - 1) = gate;

#if (defined(VERILOG))
  gate->comp_name = strdup (gate_name);
#endif
  gate_number(gate) = gate_no;
  gatefunction (gate) = p_reg;
  /*  outputindex (gate) = 0; */
  nmbrinputs (gate) = 0;
  nmbroutputs (gate) = 0;
  levelnumber (gate) = 1;
  gate_index (gate) = nmbr_reg (ckt) - 1;
  return (gate);

} /* make_new_reg */

/****************************************************************************/
/* Delete an existing wire */
/****************************************************************************/
void delete_wire (NODE *from_gate, NODE *to_gate)
{ /* delete_wire */
  int faninindex;
  int new_numoutputs, new_numinputs;
  int j, k, l, m;
  NODE * fanout;
  NODE * fanin;
  
/* Delete the wire */
  new_numoutputs = nmbroutputs (from_gate) - 1;
  new_numinputs = nmbrinputs (to_gate) - 1;
  for (j = 0; j < nmbroutputs (from_gate); ++ j)
    {
      fanout = fanoutgate (from_gate, j);
      if ((NODE *) fanout == to_gate) break;
    }
  /* j now has the port number to which to_gate connects for from_gate */
  for (k = j; k < (nmbroutputs (from_gate) - 1); ++ k)
    {
      /* Copy over fanouts that are staying on from_gate */
      fanoutgate (from_gate, k) = fanoutgate (from_gate, k + 1);
      fanoutgateinpnmbr (from_gate, k) = fanoutgateinpnmbr (from_gate, k + 1);
    }
  /* Decrement the number of outputs in from_gate's fanout list */
  nmbroutputs (from_gate) = new_numoutputs;

  if (gatefunction (to_gate) == p_po)
    {
      for (k = 0; k < (nmbroutputs (from_gate)); ++ k)
	{
	  fanout = fanoutgate (from_gate, k);
	}
    }
  /* Remove from_gate from to_gate's fanin list */
  faninindex = nmbrinputs (to_gate);
  for (k = 0; k < faninindex; ++ k)
    {
      /* if (kth fanin == from_gate) break; */
      if ((NODE *) faningate (to_gate, k) == from_gate) break;
    }
  /* k now has the port number to which from_gate connects for to_gate */
  for (l = k; l < faninindex - 1; ++ l)
    {
      /* copy fanin gatepntr of l + 1 to position l */
      /* find gate driving position l -- save it in fanin */
      fanin = (faningate (to_gate, l) = faningate (to_gate, l + 1));
      /* change port number for the fanout of driving gate to l */
      for (m = 0; m < nmbroutputs (((NODE *) fanin)); ++ m)
	{
	  if (fanoutgate (((NODE *) fanin), m) == (struct NODE *) to_gate)
	    {
	      fanoutgateinpnmbr (((NODE *) fanin), m) = l; /* Fix the port number */
	      break;
	    }
	}
    }
  nmbrinputs (to_gate) = new_numinputs;
  /*  -- outputindex (from_gate); */
  return;
} /* delete_wire */

/****************************************************************************/
/* Add a new wire */
/****************************************************************************/
void add_wire (NODE *from_gate, NODE *to_gate)
{ /* add_wire */
  int faninindex;
  int new_numoutputs, new_numinputs;
  int j;

  /* Augment the fanin list */
  if (fanins (to_gate) == NULL)
    fanins (to_gate) = (FANIN *) calloc (1, sizeof (NODE *));
  else
    fanins (to_gate) =
      (FANIN *) realloc (fanins (to_gate),
			 (1 + nmbrinputs (to_gate)) * sizeof (NODE *));
  /* Augment the fanout list */
  if (fanouts (from_gate) == NULL)
    fanouts (from_gate) = (FANOUT *) calloc (1, sizeof (FANOUT));
  else
    fanouts (from_gate) = (FANOUT *) realloc (fanouts (from_gate),
      ((1 + nmbroutputs (from_gate)) * sizeof (FANOUT)));

  /* add the wire */
  new_numoutputs = nmbroutputs (from_gate) + 1;
  new_numinputs = nmbrinputs (to_gate) + 1;

  /* put to_gate in from_gate's fanout list */
  j = nmbroutputs (from_gate);
  fanoutgate (from_gate, j) = (struct NODE *) to_gate;
#if defined (VERILOG)
  /*  fanoutgatesig (from_gate, j) = (char *) h_lookup (compname (to_gate));*/
#endif
  fanoutgateinpnmbr (from_gate, j) = new_numinputs - 1;
  nmbroutputs (from_gate) = new_numoutputs;
  /*  ++ outputindex (from_gate); */

  /* put from_gate in to_gate's fanin list */
  faninindex = nmbrinputs (to_gate);
  faningate (to_gate, faninindex) = (struct NODE *) from_gate;
  nmbrinputs (to_gate) = new_numinputs;

  return;
} /* add_wire */

