/*****************************************************************************
 * This is the top level control function for SpectralATPG.                  *
 *****************************************************************************/

#include "sim.h"
#include "all.h"

//#define OK_DEB

static int bestrdexp;
static int bestcutoff;

/* Perturb-Phase Expansion */
int pp_expandV(int **from_array, int num_bits,int num_vec, int max_shift, 
	       int **to_array, int initial_vec /*leave for future use*/)
{
  int shift, ttv;
  int i,j;

  ttv = 0; 
  for(j=0;j<num_bits;j++) 
    {
      if(max_shift==0) shift=0;
      else shift = rand()%max_shift;

      ttv = (shift>ttv) ? shift:ttv;

      if((initial_vec+num_vec+ttv) >= tmpvecsize)
	resize_tmpvec(tmpvecsize, tmpvecsize*2, num_bits);

      for(i=num_vec-1;i>=0;i--)
	{
	  (*to_array)[(i+shift+initial_vec)*num_bits+j] = (*from_array)[i*num_bits+j];
	  if((*from_array)[i*num_bits+j]!='1' && (*from_array)[i*num_bits+j]!='0')
	    fprintf(resultsout, "%c!='0','1',ttv=%d,pp_expandV\n", 
		    (*from_array)[i*num_bits+j],ttv+1);
	  if(sfsimdebug)
	    fprintf(resultsout, "%d", (*from_array)[i*num_bits+j]);
	}
      if(sfsimdebug)
	fprintf(resultsout, "\n");
      for(i=0;i<shift;i++)
	{
	  (*to_array)[(i+initial_vec)*num_bits+j] = rand()%2+'0';
	}
      for(i=num_vec+shift-1;i<num_vec+max_shift;i++)
	{
	  (*to_array)[(i+initial_vec)*num_bits+j] = rand()%2+'0';
	}
    }
  return (num_vec+ttv+initial_vec); /*this is number of the expanded vectors*/
}

int best_spectral(int iter, int totalv, int rpi, int H_size)
{
  int num_vec;
  int bestfaultcount = -1;
  int faultcount;
  int lowbound, upbound;
  int h_w = 0;
  int rdexp=0, cutoff_set;
  int bestnum_vec=0;
  int fix_hold = false;
  
  bestrdexp = 5;
  bestcutoff = 8;
  
  usebelowcutoff = 0; 
  
  /*Algorithm 2: randomly hold each vec for several clocks*/     
  // printf("entering best_spectral---\n");

  if(iter==1)
    {
      lowbound= 1; 
      upbound = 51;
    }
  else 
    {
      lowbound = ((bestrdexp-4)>1)?(bestrdexp-4):1; 
      //upbound = ((bestrdexp+5)<51)?(bestrdexp+5):51;
      upbound = ((bestrdexp+5)<(exp_iter_set+1))?(bestrdexp+5):(exp_iter_set+1);
    }
  if(iter == 0)
    {
      num_vec = cp_compacted_vec( &tmpvec1, &tmpvec2, totalv, rpi, true, 0);	
      for(cutoff_set = (H_size==2)?0:2; cutoff_set < H_size; cutoff_set += 2)
	{ 
	  /* Extend the vector sequence with the Hadamard method */
	  hadamard_expansionV(&tmpvec2, rpi, 
			      num_vec*2,h_w, cutoff_set, H_size );
	  /* Evaluate effectiveness of last Hadamard expansion */
	  faultcount = sfsim ("tmpvec.vectors", tmpvec2, nmbr_pi(circuit),
			      num_vec*2,
			      USEARRAY, resultsout, circuit, faultlist,
			      UNDETECTEDF, sfsimdebug, NCREAT_FLIST, 0,
			      NUPDATE_STATE, NUPDATE_FAULT, UNINITIAL_FAULT,
			      false,
			      0, EARLY_STOP);
	  /*printf("iter--%2d, no random hold, cutoff=%2d, Fdetected=%5d\n", 
	    iter, cutoff_set, faultcount); */
	  if( (faultcount > bestfaultcount) || 
	      ((faultcount == bestfaultcount)&&(bestnum_vec > num_vec*2)) )  
	    {
	      bestfaultcount = faultcount;
	      bestrdexp = rdexp;
	      bestcutoff = cutoff_set;
	      cp_compacted_vec( &tmpvec2, &tmpvec, num_vec*2, rpi, true, 0);
	      bestnum_vec = num_vec*2;
	    }
	}
    }
  else
    for(rdexp = lowbound; rdexp < upbound; rdexp+=2)
      {
	num_vec = random_expandV(&tmpvec1, rpi, totalv,
				 rdexp,&tmpvec2, fix_hold);
	
	for(cutoff_set = (H_size==2)?0:2; cutoff_set < 15; cutoff_set += 2)
	  { 
	    //printf("best_spectral...totalv= %d, num_vec=%d\n", totalv, num_vec );
	    /* Extend the vector sequence with the Hadamard method */
	    hadamard_expansionV(&tmpvec2, rpi, 
				num_vec*2,h_w, cutoff_set, H_size );
	    //printf("here...rdexp=%d\n", rdexp);

	    /* Evaluate effectiveness of last Hadamard expansion */
	    faultcount = sfsim ("tmpvec.vectors", tmpvec2, num_pi-num_dff,
				num_vec*2,
				USEARRAY, resultsout, circuit, faultlist,
				UNDETECTEDF, sfsimdebug, NCREAT_FLIST, 0,
				NUPDATE_STATE, NUPDATE_FAULT, UNINITIAL_FAULT,
				false,
				0, EARLY_STOP);
	    /* printf("iter--%2d, rdexp=%2d, cutoff=%2d, Fdetected=%5d\n", 
	       iter, rdexp, cutoff_set, faultcount); */
	    if( (faultcount > bestfaultcount) || 
		((faultcount == bestfaultcount)&&(bestnum_vec > num_vec*2)) )  
	      {
		bestfaultcount = faultcount;
		bestrdexp = rdexp;
		bestcutoff = cutoff_set;
		cp_compacted_vec( &tmpvec2, &tmpvec, num_vec*2, rpi, true, 0);
		bestnum_vec = num_vec*2;
	      }
	  }
      }
  printf("BEST: iter--%2d, rdexp=%2d, cutoff=%2d, Fdetected=%5d\n", 
	 iter, bestrdexp, bestcutoff, bestfaultcount); 

  // printf("exiting best_spectral---\n");
  return (bestnum_vec);
}

void perturb_H(int H_size, int iter)
     /*total_bits is the number of bits in Hmatrix*/
{
  int i, j, k;
  int *H_matrix;

  if(H_size==2)       H_matrix = (int*)(&hadamard1[0][0]);
  else if(H_size==4 ) H_matrix = (int*)(&hadamard2[0][0]);
  else if(H_size==8 ) H_matrix = (int*)(&hadamard3[0][0]);
  else if(H_size==16) H_matrix = (int*)(&hadamard4[0][0]);
  else if(H_size==32) H_matrix = (int*)(&hadamard5[0][0]);
  else if(H_size==64) H_matrix = (int*)(&hadamard6[0][0]);
  else 
    {
      printf("H_size = %d, H_size should be 2, 4, 8, 16, 32 or 64!!!--spectral()\n",H_size);
      exit(0);
    }
  j = H_size*H_size;

  if(iter != 0)
    for(i=0;i<H_size;i++)
      {
	k = rand()%j;
	*(H_matrix+k) = 0- (*(H_matrix+k));
      }

  // fprintf(resultsout, "random perturb H_matrix: Matrix(%d) = \n",iter);
  for(i=0;i<H_size;i++)
    {
      for(k=0;k<H_size;k++)
	{
	  // fprintf(resultsout,"%3d", *(H_matrix+i*H_size+k));
	}
      // fprintf(resultsout,"\n");
    }
}

void perturb_expand_V(int **tmpvec, int num_bits, int num_vec, int cycle)
     //1 bit will be fliped every 'cycle' bits
     //num_vec is the total num of vec after expansion
{
  int i, j=0, k;
  int *array;

  if(num_vec > tmpvecsize) 
    resize_tmpvec (tmpvecsize, (num_vec > tmpvecsize*2)? num_vec:(tmpvecsize*2), num_bits);
  
  k = num_vec/2;
  array = *tmpvec;
  for(i=0;i<k*num_bits;i++)
    {
      *(array+i+k*num_bits) = *(array+i);
    }
  for(i=k*num_bits;i<num_vec*num_bits-cycle;i += cycle )
    {
      j = rand()%cycle;
      *(array+i+j) = '0' + '1' - *(array+i+j);
    }
  
}
 
void randVec( int *array, int total_bits)
{
  int i, j,  k, rand_word;
  
  k=0;
  for(i=0;i<(total_bits/32);i++)
    {
      rand_word = random();
      for(j=0;j<32;j++,k++)
	array[k] = ((rand_word >> j) & 0x00000001) + '0';
    }
  for(i=(total_bits/32)*32;i<total_bits;i++)
    {
      array[i] = '0';
    }
  return;
}

int spectrum_info(char *sp_infofile, int *array, int num_vec, int num_bits, int HD_Hsize, int HD_cutoff, int HD_holding)
     /*Please note: this function will store the spectrum info in array. 
       The contant of array will be overwriten */

{
  int i,j,k;
  int half_blocksz,blocksz,num_block,tmp_coef, H_order;
  int *coef;
  FILE *ft;

  ft = fopen(sp_infofile, "w");
  coef = (int *)calloc(num_bits*HD_Hsize,sizeof(int));

  H_order = (HD_Hsize==2)?1:((HD_Hsize==4)?2:((HD_Hsize==8)?3:((HD_Hsize==16)?4:((HD_Hsize==32)?5:6))));
  if(HD_Hsize!=(int)pow(2,H_order)) 
    fprintf(ft,"HD_Hsize should be 2, 4, 8, 16, 32 or 64! --spectrum_info()\n");
  
  /* Char -> num: '0'->-1; '1'->1 */
  for(i=0;i<num_vec*num_bits;i++)
    array[i] = 2*(array[i]-'0')-1;

  /*for(i=0;i<num_vec;i++)
    {
      for(j=0;j<num_bits;j++)
	fprintf(ft, "%2d",array[i*num_bits+j]);
      fprintf(ft, "\n");
      }*/

  /* Calculate the spectral coefficients: Hadamard decomposition */
  for(i=0;i<H_order;i++)//i: current order - 1, block size = pow(2, i+1);
    {
      half_blocksz = (int)pow(2,i);
      blocksz = (int)pow(2,i+1);
      num_block = (int)(num_vec/HD_Hsize)*(HD_Hsize/blocksz);

      half_blocksz *= num_bits;//very important

      for(j=0;j<num_block;j++)//j:block #
	for(k=j*blocksz*num_bits;k<(j*blocksz*num_bits+half_blocksz);k++)//k:position in a block
	  {
	    tmp_coef = array[k]-array[k+half_blocksz];
	    array[k] = array[k]+array[k+half_blocksz];
	    array[k+half_blocksz] = tmp_coef;
	  }
      /* fprintf(ft, "H_order=%d\n", i+1);
  for(k=0;k<num_vec;k++)
    {
      for(j=0;j<num_bits;j++)
	fprintf(ft, "%2d",array[k*num_bits+j]);
      fprintf(ft, "\n");
      }*/
    }

  /*fprintf(ft, "Sp coefs:\n");
  for(i=0;i<num_vec;i++)
    {
      for(j=0;j<num_bits;j++)
	fprintf(ft, "%2d",array[i*num_bits+j]);
      fprintf(ft, "\n");
      }*/
  
  /* Filtering with HD_cutoff value: plus 1 if abs()>HD_cutoff */
  num_block = num_vec/HD_Hsize;
  blocksz = num_bits*HD_Hsize;

  for(j=0;j<num_bits;j++)
    for(k=0;k<HD_Hsize;k++)
      for(i=0;i<num_block;i++)
	if(abs(array[(i*HD_Hsize+k)*num_bits+j])>=HD_cutoff) coef[k*num_bits+j] ++;

  /* Print the spectrum information */
  fprintf(ft,"Spectral Information:\n--Size of BIST hardware Hadamard matrix = %d (value from the -U option)\n", HD_Hsize);
  fprintf(ft,"--Cutoff value of BIST Hadamard matrix-filtering operation = %d (value from the -V option)\n", HD_cutoff);
  fprintf(ft, "--Recommanded holding time = %d (Holding time need to be rounded to nearest power of 2 for implementation benefit -- Krishna )\n", HD_holding);
  fprintf(ft,"/************************************************************/\n");
  for(j=0;j<num_bits;j++)
    {
      fprintf(ft,"The counter values of PI #%d:\n",j+1);
      for(k=0;k<HD_Hsize;k++)
	fprintf(ft,"%d\n", coef[k*num_bits+j]);
    }

  free(coef);
  fclose(ft);
  return 1;
} 

int main (int argc, char *argv[])
{ /* main */
  int		error_flag;
                           /* num_vectors is the number of vectors after
			      each iteration */
                           /* detected_flts is total number of detected
			      faults at this algorithm stage */
  int allnv, bestallnv;    /* Number of vectors (random or extension)
			      generated before compaction to
			      get final result */
  int totalt;              /* Total CPU time in sec. */
  int faultcount;         /* Number of faults this fault sim. pass got --
			      this is not the total */
  int totalv;              /* Total vector length at this algorithm stage */
  int i=0, m=0;              /* Loop indices */ 
  int removed_vec=0;       /* Number of vectors removed by both types of
			      compacter in the last operation */
  char finalvecfile[40];   /* Name of final vector file */
  char sp_infofile[40];    /* Name of file to store BIST spectral information of test set*/
  FILE *ft;                /* File pointer to final vector file */
  int h_w = 0;             /* Indicates which transform to use for extension:
			      0: Hadamard expansion 
			      1: Walsh
			      2: 
			      3: */
  int rdexp = 9;           /* For random_expand() -- maximum number of times
			      each vector can be held */
  int max_sp_iter=125;     /* Number of Hadamard expansion iterations */
  int initnv=0;            /* Initial number of vectors before Hadamard
			      expansion */
  int initndet=0;          /* Initial number of faults detected by random
			      pattern generation */
  int iter=0;              /* Total number of Hadamard expansion iterations */
  char *commonresults = "commonresults"; /* Name of common result file */
  int cutoff_set = 8;
  int fix_hold = false; //true: hold for fixed time; false: random
  int H_size = 16;
  int cycle[8]={2,4,6,8,10,12,14,16};//{2,4,6,8};
  int five_set = 1, five_iter=0, tech_num=0, tech_iter;
  int SG_iter=0, bit_ptb_iter=30, noisyM_iter=0, phase_iter=0;
  int HD_Hsize = 8, HD_cutoff = 4; /* For extracting the BIST coefficients */
  int HD_holding = 1;
  int atpg_num = 0; /* The matrix size or perturbation period in */
  int ptb_period, fs_shift;

  int gate; /* OK */
  int index;/* OK */
  int ok_debugging; /* OK */
  int ppo_cntr = 0;
#define SG_VEC_NUM 16 /* No. of vectors to be randomly perturbed */
#define INIT_VEC_N 40000 /*the size of tmpvec and tmpvec1*/

  /*Set not printing the vectors and good machine responses by sfsim()*/
  prt_VEC_GM = 0; 

  elapsetime_s(resultsout,TIME_TOTAL,false);

  creat_D_state = 0;
  inv_transform = 1; //with inverse transformation
  no_filter = 0; //use the cutoff value to filter
  flip_bit = 0; //don't flip any bit 

  sg_malloc = false; 
  init_main_fsim ();
  strinit ();  

  /* Get command line options */
  setoptions(argc,argv);

  elapsetime_s(resultsout,TIME_PARSE,false);

  /* Parse the input circuit hardware description language and fault list */
  error_flag = r_parse (&circuit, sch_file_name, 1);
  if (error_flag < 0)
    {
      printf ("parsing failed\n");
      exit (-1);
    }
  else if (error_flag == 0)
    {
      printf ("Trying to do sequential ATPG on a combinational circuit.\n");
    }
  readfaults(/*"s298.faults"*/fault_file_name);/* Fix so it accepts files from coomand line */
  /* Allocate memory*/
  tmpvecsize = INIT_VEC_N;
  tmpvec  = (int *)calloc((num_pi-num_dff)*tmpvecsize, sizeof(int));
  tmpvec1 = (int *)calloc((num_pi-num_dff)*tmpvecsize, sizeof(int)); 
  tmpvec2 = (int *)calloc((num_pi-num_dff)*tmpvecsize, sizeof(int)); 
  removed = (int *)calloc(tmpvecsize, sizeof(int));
  removed_tmp = (int *)calloc(tmpvecsize, sizeof(int));
  alldflist = (int*)calloc(faultcnt+1,sizeof(int));
  alldflist[0] = 0;

  elapsetime_e(resultsout,TIME_PARSE,false);  

  sfsimdebug = false;

  //printf("HD_Hsize_set = %d, HD_cutoff_set = %d\n", HD_Hsize_set, HD_cutoff_set);
  //if(exp_iter_set != -1) max_sp_iter = exp_iter_set; 
  if(exp_iter_set == -1) exp_iter_set=64; // now exp_iter_set is used to set holding time
  if(Hsize_set != -1 && (Hsize_set-4)*(Hsize_set-8)*(Hsize_set-16)*(Hsize_set-32)*(Hsize_set-64)==0 )
    H_size = Hsize_set;
  else if(Hsize_set != -1)
    printf("Hsize is set to %d, should be 2, 4, 8, 16, 32 or 64!\n",Hsize_set);
  if(HD_Hsize_set != -1 && (HD_Hsize_set-4)*(HD_Hsize_set-8)*(HD_Hsize_set-16)*(HD_Hsize_set-32)*(HD_Hsize_set-64)==0 )
    HD_Hsize = HD_Hsize_set;
  else if(HD_Hsize_set != -1)
    printf("HD_Hsize is set to %d, should be 2, 4, 8, 16, 32 or 64!\n",HD_Hsize_set);
  if(HD_cutoff_set != -1) HD_cutoff = HD_cutoff_set;
 
  if(SG_set != -1 || bit_ptb_set != -1 || noisyM_set != -1 || phase_set != -1) 
    SG_iter = bit_ptb_iter = noisyM_iter = phase_iter = 0;
  if(SG_set != -1) SG_iter = SG_set;
  if(bit_ptb_set != -1) bit_ptb_iter = bit_ptb_set;
  if(noisyM_set != -1) noisyM_iter = noisyM_set;
  if(phase_set != -1) phase_iter = phase_set;


  sprintf(sp_infofile, "%s%s", circuit_name, ".BIST_coef");
  if(use_timeflag == true)
    sprintf(finalvecfile, "%s%c%s%s", circuit_name, '_',
	    time_flagg,".vectors");
  else
    sprintf(finalvecfile, "%s%s", circuit_name,".vectors");

  //Okhan New Code

  totalv = initnv = allnv = bestallnv = 0;
  detected_flts = new_detected_flts = initndet = faultcount = 0;

  num_vectors = creat_tmp_vecfile(finalvecfile,"tempVec");
  //fprintf(stdout, "Calling fsim: output will be written: #vec=%d, #f_det=%d\n"
  //	  ,num_vectors, faultcount);

  /* WARNING -- Not sure about the range of the 'for' loop, please recheck!! -- OK */
  for (index=1;index<=gatecnt-num_dff;index++)
    {
      if((get_node_pntr(circuit, index))->funct == p_pi)
	{
	}
      //if (inputlist(index)!=NULL)
      else
	{
	  if (fanins(get_node_pntr(circuit, index))!=NULL)
	    {
	      i = nmbrinputs(get_node_pntr(circuit, index));
	      ((get_node_pntr(circuit, index))->lines[0]) = (int *)calloc(i+1, sizeof(int));
	      //circuit[index]->lines[1] = (int *)make_ilist(i);
	      ((get_node_pntr(circuit, index))->lines[1]) = (int *)calloc(i+1, sizeof(int));
	      //ilist_length(circuit[index]->lines[0]) = i;
	      //ilist_length(((get_node_pntr(circuit, index)))->lines[0]) = i;
	      //ilist_length(circuit[index]->lines[1]) = i;
	      //ilist_length(((get_node_pntr(circuit, index)))->lines[1]) = i;
	    }
	}
    }
  /* Allocating Memory for lines in PPOs */
  for (index=0; index<num_dff; index++)
    {
      PPOs(circuit, index)->lines[0] = (int *)calloc(2, sizeof(int));
      PPOs(circuit, index)->lines[1] = (int *)calloc(2, sizeof(int));
    }
  /* Verifying Circuit Connections -- OK */
#ifdef OK_DEB
  ok_debugging = print_ckt_topology(circuit, gatecnt, nmbr_pi(circuit), nmbr_po(circuit));
  ok_debugging = print_lev_ckt_topology(circuit, gatecnt, nmbr_pi(circuit), nmbr_po(circuit));
  if(ok_debugging > 0)
    printf("Justified Circuit Netlist! Its a Go!\n");
  else
    printf("WARNING -- Something WRONG!\n");

  printf("Calling fsim: output will be written: #vec=%d, #f_det=%d\n"
	 ,num_vectors, faultcount);
  /* Establish a baseline fault coverage */
  detected_flts = sfsim (finalvecfile, tmpvec, nmbr_pi (circuit), 
		      num_vectors,
		      USEFILE, resultsout, circuit, faultlist,
		      UNDETECTEDF, sfsimdebug, NCREAT_FLIST, 0,
		      UPDATE_STATE, UPDATE_FAULT, INITIAL_FAULT, false/*'not compacted'*/,
		      0, EARLY_STOP);
  fprintf(stdout, "Finished Simulation: output written to %s: #vec=%d, #f_det=%d\n",
	  resultsout,num_vectors, detected_flts);
#endif

  //end Okhan new code

  
  if(1)
    {  //Okhan  
  /* Empty out the finalvecfile */
  ft =  fopen(finalvecfile, "w");
  fclose(ft);
  /************************************************************************************/ 
  /*Start with random vectors*/
  if(1)
    {
      ft =  fopen(vectorsfile_name, "w");
      fclose(ft);
      totalv = initnv = allnv = bestallnv = 0;
      detected_flts = initndet = faultcount = 0;
      for(i=0,five_iter=0;i<10&&five_iter<3;i++)
	{
	  /* Generate random sequences */
	  randVec(tmpvec,(nmbr_pi (circuit)*500));
	  /* Use LROR compaction */
	  num_vectors = 500 - rst_compaction(tmpvec,500,nmbr_pi (circuit));
	  /* Append the compacted sequence to tmpvec1, which stores all the useful vectors */
	  totalv += cp_compacted_vec(&tmpvec,&tmpvec1,500 ,
				     nmbr_pi (circuit), false/*means 'compacted'*/, totalv);
	  allnv += 500;
	  /* Establish a baseline fault coverage */
	  faultcount = sfsim ("tmpvec.vectors", tmpvec, nmbr_pi (circuit), 
			      num_vectors,
			      USEARRAY, resultsout, circuit, faultlist,
			      UNDETECTEDF, sfsimdebug, NCREAT_FLIST, 0,
			      UPDATE_STATE, UPDATE_FAULT, UNINITIAL_FAULT, true/*'compacted'*/,
			      0, EARLY_STOP);
	  detected_flts += faultcount;
	  initndet += faultcount;
	  five_iter = (faultcount == 0)? five_iter+1:0;
	}
      initnv = totalv;
      bestallnv = allnv;
      /* Backup the compacted vectors to a file like s27.vectors */
      append_vec(&tmpvec1,vectorsfile_name,totalv,nmbr_pi (circuit),true/*means 'append each vec'*/,
		 false/*means 'overwrite (instead of append)'*/);
    }//okhan_stuff
  
  /************************************************************************************/ 
  /*seqence expansion iterations*/	
  initialize_hadamard(0);
  five_iter = 0;
  bestallnv =0;
  m=initndet; 
  iter=0;
  totalt=0;
  
  fprintf(resultsout,"\nspectralatpg.sun4:\n%37s#Compacted vec---#F_det---F_Coverg---Note\n", 
	  "Iteration------------------------------");
  
  fprintf(resultsout,"%43s   %7d   %5d   %7.4f %c\n", 
	  "Initial Sequence:",totalv,
	  detected_flts,(float) (detected_flts) * 100.0 / faultcnt, '%');
  fflush(resultsout);
  
  /************************************************************************************/ 
  /*seqence expansion iterations*/	
  for(i=0,five_iter=0,tech_num=1,tech_iter=0;(i<max_sp_iter) && (tech_num<5) && (detected_flts<faultcnt); i++,tech_iter++)
    { 
      five_set = 1;
      if(tech_num==1 && bit_ptb_iter==0) tech_num ++;
      if(tech_num==2 && noisyM_iter==0) tech_num ++;
      if(tech_num==3 && phase_iter==0) tech_num ++;
      if(tech_num==4 && SG_iter==0) tech_num ++;
 
      if(tech_num==5) break;
      else
	{
	  /* Evolve for the best cutoff value, holdtime and spectra 
	     Expand the vectors in tmpvec1 (totalv) and store in tmpvec */
	  if(tech_num==4) 
	    {
	      num_vectors = best_spectral(tech_iter, totalv, nmbr_pi (circuit), H_size);
	      HD_holding = bestrdexp;
	      atpg_num = H_size;

	      /* Run the selfish gene algorithm */
	      if(num_vectors > SG_VEC_NUM)
		selfish_gene_spectral(SG_VEC_NUM, nmbr_pi (circuit), tmpvec, num_vectors, 50, 2);
	      printf("%2d  Spectral with SG Algorithm2: #New vec=%d\n",i+1, num_vectors);
	      fprintf(resultsout,"%2d  Spectral with SG Algorithm2: #New vec=%d\n",i+1, num_vectors);
	      if( tech_iter == SG_iter-1) 
		{
		  tech_num ++;
		  initialize_hadamard(0);
		  five_set = 0;
		  tech_iter = -1;
		}
	    }	  
	  if((tech_num==1) || (tech_num==2) || (tech_num==3))
	    {
	      /*to use the weak spectral*/
	      usebelowcutoff = 0; //(i/4)%2;	  
	      rdexp = (i)%exp_iter_set+1;
	      fix_hold = true;
	      H_size = 16;
	      num_vectors = totalv;
	      
	      /*Algorithm 2: randomly hold each vec for several clocks,
		The extended vectors are stored in tmpvec*/ 
	      if(tech_iter>0)
		num_vectors = random_expandV(&tmpvec1,nmbr_pi (circuit),totalv,
					     rdexp,&tmpvec, fix_hold);
	      /*Phase_Perturbation expansion*/
	      if(tech_num==3)
		{
		  if(fs_shift_set != -1) fs_shift = fs_shift_set;
		  else fs_shift = cycle[tech_iter%8];
		  atpg_num = fs_shift;

		  num_vectors = pp_expandV(&tmpvec,nmbr_pi (circuit),num_vectors,fs_shift,
					   /*2*macro_iter,*/ &tmpvec,num_vectors);
		  printf("%2d  Phase Perturbation: #New vec=%d\n",i+1, num_vectors);
		  fprintf(resultsout,"%2d  Phase Perturbation: #New vec=%d\n",i+1, num_vectors);
		  if( tech_iter == phase_iter-1) 
		    {
		      tech_num ++;
		      initialize_hadamard(0);
		      five_set = 0;
		      tech_iter = -1;
		    }
		}
	      /*The noisy_Hadamard matrix method */
	      if(tech_num==2) 
		{
		  num_vectors = num_vectors*2; 
		  perturb_H(H_size,tech_iter);
		  hadamard_expansionV(&tmpvec, nmbr_pi (circuit), 
				  num_vectors,h_w,(cutoff_set=(i%5)*2+4), H_size);
		  atpg_num = H_size;
		  printf("%2d  Noisy Matrix Method: #New vec=%d\n",i+1, num_vectors);
		  fprintf(resultsout,"%2d  Noisy Matrix Method: #New vec=%d\n",i+1, num_vectors);
		  if( tech_iter == noisyM_iter-1) 
		    {
		      tech_num ++;
		      initialize_hadamard(0);
		      five_set = 0;
		      tech_iter = -1;
		    }
		}
	      /*Random bit_perturbation expansion by flipping 1/cycle of the bits*/
	      if(tech_num == 1) 
		{
		  if(Ptb_period_set != -1) ptb_period = Ptb_period_set;
		  else ptb_period = cycle[tech_iter%8];
		  atpg_num = ptb_period;

		  num_vectors = num_vectors*2; 
		  perturb_expand_V(&tmpvec,nmbr_pi (circuit),num_vectors,ptb_period);

		  printf("%2d  Random Bit-Perturbation: #New vec=%d\n",i+1, num_vectors);
		  fprintf(resultsout,"%2d  Random Bit-Perturbation: #New vec=%d\n",i+1, num_vectors);
		  if( tech_iter == bit_ptb_iter-1) 
		    {
		      tech_num ++;
		      initialize_hadamard(0);
		      five_set = 0;
		      tech_iter = -1;
		    }
		}
	    }
	  allnv += num_vectors;
	  
	  /*******************************************************************************/
	  /* Use LROR compaction */
	  removed_vec = rst_compaction(tmpvec,num_vectors,nmbr_pi (circuit));
	  
	  if(sfsimdebug) fprintf(resultsout, "rst-cmp:num_vectors = %d, removed_vec=%d\n",
				 num_vectors, removed_vec);
	  
	  /*append all the compacted vectors to tmpvec1, which stores all the useful vectors*/
	  totalv += cp_compacted_vec(&tmpvec,&tmpvec1,num_vectors,
				     nmbr_pi (circuit), false, totalv);
	  /* Correct number of vectors */    
	  num_vectors = num_vectors - removed_vec;
	  /* Validate that the fault coverage is correct */
	  faultcount = sfsim ("tmpvec.vectors", tmpvec, nmbr_pi (circuit),
			      num_vectors,
			      USEARRAY, resultsout, circuit, faultlist,
			      UNDETECTEDF, sfsimdebug, NCREAT_FLIST, 0, 
			      UPDATE_STATE, UPDATE_FAULT, UNINITIAL_FAULT,
			      true,
			      0, EARLY_STOP);
	  
	  /*count 5 iterations if there is not new faults detected*/
	  if((faultcount != 0) || (five_set==0) ) five_iter = 0;
	  else five_iter ++;
	  if(five_iter == 5) 
	    {
	      tech_num ++;
	      initialize_hadamard(0);
	      five_iter = 0;
	      tech_iter = -1;
	    }
	  
	  detected_flts += faultcount;
	  
	  if(detected_flts > m)
	    {
	      m = detected_flts;
	      iter = i+1;
	      totalt = elapsetime_m(resultsout,TIME_TOTAL,false);
	      bestallnv = allnv;
	      HD_holding = (HD_holding>rdexp)?HD_holding:rdexp;
	    }	  

	  /* if the following statement is used, iter will be number of last iteration*/
	  iter = i+1;
	  
	  fprintf(resultsout,"%43s   %7d   %5d   %7.4f %c%7d\n",
		  "After Compaction (LROR):",totalv,
		  detected_flts, (float) (detected_flts) * 100.0 / faultcnt,
		  '%', atpg_num);
	  printf("%43s%9d%9d%10.4f %c\n",
		 "After Compaction (LROR):",totalv,
		 detected_flts, (float) (detected_flts) * 100.0 / faultcnt,
		 '%');
	  fflush(resultsout);
	  
	  if((i%3)==0)
	    {
	      totalv = append_vec(&tmpvec1,finalvecfile,totalv,nmbr_pi (circuit),true,false);
	      ft = fopen(finalvecfile, "a");
	      fprintf(ft, "iter=%d, fault_det=%d\n",i,detected_flts);
	      fclose(ft);
	    }
	}//else..
    }//for(i=0...
  /*************************************************************************************/
  // Final verification of the entire vector sequence
  sfsimdebug = VERBOSE;
  if(sfsimdebug) fprintf(resultsout, "\nFaults being detected:\n");
  detected_flts = sfsim (finalvecfile, tmpvec1, nmbr_pi (circuit), totalv,
			 USEARRAY, resultsout, circuit, faultlist,
			 UNDETECTEDF, sfsimdebug, NCREAT_FLIST, 0, 
			 NUPDATE_STATE, UPDATE_FAULT, INITIAL_FAULT, false,
			 0, EARLY_STOP);
  fprintf(resultsout,"\n%43s%10d%8d%10.4f %c\n",
	  "Final Verification:", totalv,
	  detected_flts, (float) (detected_flts) * 100.0 / faultcnt, '%');
  fprintf(resultsout,"\n\n***** Meaning of 'Note' ******\n");
  fprintf(resultsout,"For Random Bit-Perturbation: period of bit perturbation\n");
  fprintf(resultsout,"For Noisy Matrix Method: size of the noisy matrix\n");
  fprintf(resultsout,"For Phase Perturbation: # of maximum cycles of phase perturbation\n");
  fprintf(resultsout,"For Spectral with SG Algorithm 2: size of Hadamard matrix\n");
  
  /*************************************************************************************/
  if(0){
    // Final compaction of the entire vector sequence
    removed_vec = oms_compaction(tmpvec1,totalv,nmbr_pi (circuit));
    
    sfsimdebug = 0;//true;
    detected_flts =sfsim (finalvecfile, tmpvec, nmbr_pi (circuit), totalv,
			  USEFILE, resultsout, circuit, faultlist,
			  UNDETECTEDF, sfsimdebug, NCREAT_FLIST, 0, 
			  NUPDATE_STATE, UPDATE_FAULT, INITIAL_FAULT, false,
			  0, NEARLY_STOP);
    fprintf(resultsout,"%43s   %7d   %5d   %7.4f %c\n",
	    "After Final Cmpct:",totalv,
	    detected_flts, (float) (detected_flts) * 100.0 / faultcnt, '%');
    sfsimdebug = 0;
  }
  
  /*************************************************************************************/
  /* Output stage */
  totalv = append_vec(&tmpvec1, finalvecfile, totalv, nmbr_pi (circuit), true, false);

  ft = fopen(commonresults, "a");
  fprintf(ft, "Ckt_Name Total#_flts  #vectors  #detected_flts  #iterations  CPUtime(sec)\n");
  fprintf(ft, "  %6s   %9d   %7d   %13d   %10d   %11d\n", 
	  circuit_name, faultcnt, totalv, detected_flts, iter, totalt);  
  fclose(ft);

    } //Okhan if(0)
  
  //reset_time(TIME_TOTAL);
  
  if (VERBOSE == TRUE)
    {
      fprintf(resultsout, "\nFaults not being detected:\n");
      for(faultlist_index=1;faultlist_index<=faultcnt;faultlist_index++)
	{
	  if(faultlist[faultlist_index]->flag == false)
	    fprintf(resultsout,"%d %d %d\n",
		    faultlist[faultlist_index]->from,
		    faultlist[faultlist_index]->to,
		    faultlist[faultlist_index]->sa);
	}
    }
  
  /* To print the vectors and good machine responses into finalvecfile */
  prt_VEC_GM = 0;  
if(prt_VEC_GM) detected_flts =sfsim (finalvecfile, tmpvec1, nmbr_pi (circuit), totalv,
			USEARRAY, resultsout, circuit, faultlist,
			UNDETECTEDF, sfsimdebug, NCREAT_FLIST, 0, 
			NUPDATE_STATE, NUPDATE_FAULT, INITIAL_FAULT, false,
			0, NEARLY_STOP);

  //totalv = get_vec("s1488.vectors", 8, 1, 10000);
  spectrum_info(sp_infofile, tmpvec1, totalv, nmbr_pi (circuit), HD_Hsize, HD_cutoff, HD_holding);
  
  elapsetime_e(resultsout,TIME_TOTAL,false);
  /* Used to print results */
  printresults(circuit_name,resultsout, detected_flts, faultcnt);
  
  fclose(resultsout);
  
  exit(0);
  return 0;
} /* main */

int print_ckt_topology(CIRCUIT *ckt, int gatecnt, int nmbr_pi, int nmbr_po)
{
  int i,j;

  for(i=0;i<nmbr_pi; i++)
    {
      printf("Gate %d -> ", i+1);
      for(j=0; j<PIs(ckt, i)->nmbr_outputs; j++)
	{
	  printf("%d Type %s", fanoutgatenmbr(PIs(ckt, i), j), r_trans(fanoutgate(PIs(ckt, i), j)->funct));
	}
      printf("\n");
    }
  for(i=1;i<=nmbr_gates(ckt); i++)
    {
      printf("Gate %d <- ", gatesx_dup(ckt,i-1)->gate_nmbr);
      for(j=0; j<nmbrinputs(ckt->Gate_list_dup[i-1]); j++)
	{
	  printf("%d Type %s", faningate(gatesx_dup(ckt,i-1), j)->gate_nmbr, r_trans(faningate(gatesx_dup(ckt, i-1), j)->funct));
	}
      printf("\n");
      for(j=0; j<nmbroutputs(gatesx_dup(ckt, i-1)); j++)
	{
	  printf("%d Type %s", fanoutgate(ckt->Gate_list_dup[i-1], j)->gate_nmbr, r_trans(fanoutgate(gatesx_dup(ckt, i-1), j)->funct));
	}
      printf("\n");
    }
  for(i=0;i<nmbr_po; i++)
    {
      printf("PO %d -> ", i+1+nmbr_pi+nmbr_gates(ckt));
      for(j=0; j<POs(ckt, i)->nmbr_inputs; j++)
	{
	  printf("%d Type %s", faningate(POs(ckt, i), j)->gate_nmbr, r_trans(faningate(POs(ckt, i), j)->funct));
	}
      printf("\n");
    }
  return 1;
}

int print_lev_ckt_topology(CIRCUIT *ckt, int gatecnt, int nmbr_pi, int nmbr_po)
{
  int i,j;

  printf("\n\nLevelized ckt:\n");
  for(i=0;i<nmbr_pi; i++)
    {
      printf("Gate %d -> ", i+1);
      for(j=0; j<PIs(ckt, i)->nmbr_outputs; j++)
	{
	  printf("%d Type %s", fanoutgatenmbr(PIs(ckt, i), j), r_trans(fanoutgate(PIs(ckt, i), j)->funct));
	}
      printf("\n");
    }
  for(i=1;i<=nmbr_gates(ckt); i++)
    {
      printf("Gate %d <- ", gatesx(ckt,i-1) -> gate_nmbr);
      for(j=0; j<nmbrinputs(ckt->Gate_list[i-1]); j++)
	{
	  printf("%d Type %s", faningate(gatesx(ckt,i-1), j)->gate_nmbr, r_trans(faningate(gatesx(ckt, i-1), j)->funct));
	}
      printf("\n");
      for(j=0; j<nmbroutputs(gatesx(ckt, i-1)); j++)
	{
	  printf("%d Type %s", fanoutgate(ckt->Gate_list[i-1], j)->gate_nmbr, r_trans(fanoutgate(gatesx(ckt, i-1), j)->funct));
	}
      printf("\n");
    }
  for(i=0;i<nmbr_po; i++)
    {
      printf("PO %d -> ", i+1+nmbr_pi+nmbr_gates(ckt));
      for(j=0; j<POs(ckt, i)->nmbr_inputs; j++)
	{
	  printf("%d Type %s", faningate(POs(ckt, i), j)->gate_nmbr, r_trans(faningate(POs(ckt, i), j)->funct));
	}
      printf("\n");
    }
  return 1;
}
