//**********************************************************************************
// 'matrmult'  
//**********************************************************************************
//#include <stdio.h>
//#include <stdlib.h>
//#include <string.h>
#include "all.h"

int walsh3[8][8]    = {{ 1, 1, 1, 1, 1, 1, 1, 1},
		       { 1, 1, 1, 1,-1,-1,-1,-1},
		       { 1, 1,-1,-1,-1,-1, 1, 1},
		       { 1, 1,-1,-1, 1, 1,-1,-1},
		       { 1,-1,-1, 1, 1,-1,-1, 1},
		       { 1,-1,-1, 1,-1, 1, 1,-1},
		       { 1,-1, 1,-1,-1, 1,-1, 1},
		       { 1,-1, 1,-1, 1,-1, 1,-1}};

int rad_walsh3[8][8]= {{ 1, 1, 1, 1, 1, 1, 1, 1},
		       { 1, 1, 1, 1,-1,-1,-1,-1},
		       { 1, 1,-1,-1, 1, 1,-1,-1},
		       { 1,-1, 1,-1, 1,-1, 1,-1},
		       { 1, 1,-1,-1,-1,-1, 1, 1},
		       { 1,-1, 1,-1,-1, 1,-1, 1},
		       { 1,-1,-1, 1, 1,-1,-1, 1},
		       { 1,-1,-1, 1,-1, 1, 1,-1}};

int walsh_paley3[8][8]=
                      {{ 1, 1, 1, 1, 1, 1, 1, 1},
		       { 1, 1, 1, 1,-1,-1,-1,-1},
		       { 1, 1,-1,-1, 1, 1,-1,-1},
		       { 1, 1,-1,-1,-1,-1, 1, 1},
		       { 1,-1, 1,-1, 1,-1, 1,-1},
		       { 1,-1, 1,-1,-1, 1,-1, 1},
		       { 1,-1,-1, 1, 1,-1,-1, 1},
		       { 1,-1,-1, 1,-1, 1, 1,-1}};


//#include "Include/all.h"
void hadamard_expansion(char *fromfile_name, char *tofile_name, int num_pi, int num_vec, 
			int H_W, int cutoff_set, int H_size )
     /*H_W: HARDAMARD or WALSH*/
 
{
  char line [3001] = "";
  int i=0,j=0;
  int *vec;
  FILE *fromfp, *tofp;

  //initialize_hadamard(H_W);
  fromfp = fopen(fromfile_name, "r");
  
  vec = (int *)calloc(num_vec*num_pi,sizeof(int));

  while ((fgets (line, 3000, fromfp) != NULL) && (i<num_vec/2))
    {
      for(j=0;j<num_pi;j++)
	vec[i*num_pi+j] = line[j];
      i++;
    }


  //i is now the number of vectors read from the fromfp
  if(i<num_vec) spectral( i, num_pi, num_vec-i, &vec, cutoff_set, H_size);

  fclose(fromfp);
  tofp   = fopen(tofile_name, "w");

  for(i=0;i<num_vec;i++)
    {
      for(j=0;j<num_pi;j++) fprintf(tofp, "%c",vec[i*num_pi+j]);
      fprintf(tofp, "\n");
    }
  fclose(tofp);
}

void hadamard_expansionV(int **array,int num_pi, int num_vec, int H_W, int cutoff_set, int H_size )
     /*H_W: HARDAMARD or WALSH*/
{
  // initialize_hadamard(H_W);
  
  if(num_vec > tmpvecsize) 
      resize_tmpvec (tmpvecsize, (num_vec > tmpvecsize*2)? num_vec:(tmpvecsize*2), num_pi);

  //num_vec/2 is now the number of initial vectors
  spectral( num_vec/2, num_pi, num_vec/2, array, cutoff_set, H_size);

}

void initialize_hadamard(int hw)
{
  int i, j;

  /* Generating Hadamard Matrice */
  hadamard1[0][0]=hadamard1[0][1]=hadamard1[1][0]=1;
  hadamard1[1][1]=-1;

  for(i=0;i<4;i++)
    for(j=0;j<4;j++)
      {
	if(i>=2 && j>=2) hadamard2[i][j] = -hadamard1[i%2][j%2];
	else hadamard2[i][j] = hadamard1[i%2][j%2];
      }
  for(i=0;i<8;i++)
   { for(j=0;j<8;j++)
      {
	if(hw == 0) //hadamard
	  {
	    if(i>=4 && j>=4) hadamard3[i][j] = -hadamard2[i%4][j%4];
	    else hadamard3[i][j] = hadamard2[i%4][j%4];
	  }
	else if(hw == 1) //walsh 
	  hadamard3[i][j] = walsh3[i][j];
	else if(hw == 2) //rademacher_walsh 
	  hadamard3[i][j] = rad_walsh3[i][j];
	else if(hw == 3) //walsh_paley 
	  hadamard3[i][j] = walsh_paley3[i][j];
	else  //wrong hw
	  printf("no corresponding transformation!!!!\n\n\n\n");
	//	printf("%3d",hadamard3[i][j]);
      }
   // printf("\n");
   }
  //  printf("H_W = %d\n\n",hw);
  
  for(i=0;i<16;i++)
    for(j=0;j<16;j++)
      {
	if(i>=8 && j>=8) hadamard4[i][j] = -hadamard3[i%8][j%8];
	else hadamard4[i][j] = hadamard3[i%8][j%8];
      }
  for(i=0;i<32;i++)
    for(j=0;j<32;j++)
      {
	if(i>=16 && j>=16) hadamard5[i][j] = -hadamard4[i%16][j%16];
	else hadamard5[i][j] = hadamard4[i%16][j%16];
      }
  for(i=0;i<64;i++)
    for(j=0;j<64;j++)
      {
	if(i>=32 && j>=32) hadamard6[i][j] = -hadamard5[i%32][j%32];
	else hadamard6[i][j] = hadamard5[i%32][j%32];
      }
}

int MatrMult(int m, int t, int n, int *M1, int *M2, int *M3)
{
  int i, j, k;  

  for(i=0;i<m;i++)
    for(j=0;j<n;j++)
      {
	(M3)[i*n+j]=0;
        for(k=0;k<t;k++)
	   (M3)[i*n+j] = (M3)[i*n+j] + (M1)[i*t+k] * (M2)[k*n+j];
      }
  return 1;
}

void test_cutoff()
{
  int i,j,k;
  int tmpVect16[16][1],coef116[1][16];
  int coefbit[33];
  FILE *ft;

  ft = fopen("hadam_coef", "w");
  initialize_hadamard(0);
  for(j=0;j<33;j++)
    coefbit[j] = 0; 

  for(i=0; i<65536; i++)
    {
      for(j=0;j<16;j++)
	{
	  tmpVect16[j][0]=2 * ((i>>j)&01) - 1;//change 0 to -1
	}
      MatrMult(16,16,1,(int*)hadamard4,(int*)tmpVect16,(int*)coef116);//get corelation coef
      for(k=15;k>=0;k--)
	{
	  fprintf(ft, "%2d-%3d ", tmpVect16[k][0], coef116[0][k]);
	  coefbit[(coef116[0][k]+16)] += 1;
	} 
      fprintf(ft, "\n");
    }
  fprintf(ft, "******summary******\n");
  for(j=0;j<33;j++)
    fprintf(ft, "%2d--%d\n",j, coefbit[j]);

  fclose(ft); 
}

int spectral(int init_vec, int num_bits, int new_vec, int **vec, int cutoff_set, int H_size) 
     //init_vec: initial length of the test sequence
     //num_bits: number of bits in one vector
     //new_vec:  number of new vectors to be expanded
     //(*vec)[init_vec+new_vec][num_bits]
{
  int i,j,start_vec,k;
  int tmpVect[64],coef[64];
  int cutoff = 4;
  int row;
  int *H_matrix;
  //int H_size;

  if(cutoff_set != 0)
    cutoff = cutoff_set;

  //  printf("&h3=%d,&(h3[0])=%d,h3[0]=%d\n", (int)(&hadamard3),(int)&(hadamard3[0]),(int)&(hadamard3[0][0]));
  //printf("*(&h3)=%d,*(&(h3[0]))=%d,h3[0][0]=%d\n", (int)*(&hadamard3),(int)*(*((int**)(&(hadamard3[0])))+63),hadamard3[7][7]);
  fflush(stdout);

  /*select proper transformation matrix*/
  if(H_size==2)       H_matrix = (int*)(&hadamard1[0][0]);
  else if(H_size==4 ) H_matrix = (int*)(&hadamard2[0][0]);
  else if(H_size==8 ) H_matrix = (int*)(&hadamard3[0][0]);
  else if(H_size==16) H_matrix = (int*)(&hadamard4[0][0]);
  else if(H_size==32) H_matrix = (int*)(&hadamard5[0][0]);
  else if(H_size==64) H_matrix = (int*)(&hadamard6[0][0]);
  else 
    {
      printf("H_size = %d, H_size should be 2, 4, 8, 16, 32 or 64!!!--spectral()\n",H_size);
      exit(0);
    }

 if(0) for(i=0;i<H_size;i++)
    {
      for(j=0;j<H_size;j++)
	printf("%3d",(H_matrix)[i*H_size+j]);
      printf("\n");
    }

  j = (init_vec<new_vec) ? init_vec:new_vec;

  for(row=0;row<(int)(j/H_size);row++)
    for(i=0;i<num_bits;i++)
      {
	start_vec = init_vec+row*H_size;

	/*Extracting Hadamard Coefficient*/
	for(k=0;k<H_size;k++)
	  {
	    tmpVect[k]=2*((*vec)[(start_vec-init_vec+k)*num_bits+i]-'0')-1;//change 0 to -1
	  }
	MatrMult(H_size,H_size,1,H_matrix,tmpVect,coef);//get correlation coef

	/*filter or no filter on the correlation coefs*/
	for(k=0;k<H_size;k++)
	  {
	    if(no_filter == 1 && inv_transform == 0)
	      {
		if(coef[k] > 0) 
		  (*vec)[(start_vec+k)*num_bits+i]='1';
		else 
		  (*vec)[(start_vec+k)*num_bits+i]='0';
	      }
	    /*to use coefs greater than cutoff*/
	    if(no_filter == 0)
	      {
		if(usebelowcutoff == 0)
		  {
		    if(coef[k]>(-1*cutoff) && coef[k]<cutoff) 
		      {
			(*vec)[(start_vec+k)*num_bits+i]='0';
			coef[k]=0;
			/*printf("filtered!!\n\n");*/
		      }
		    else (*vec)[(start_vec+k)*num_bits+i]='1';
		  } 
		else //use coef less than cutoff
		  { 
		    if(coef[k]<(-1*cutoff) && coef[k]>cutoff) 
		      {
			(*vec)[(start_vec+k)*num_bits+i]='1';
			coef[k]=0;
			/*printf("filtered!!\n\n");*/
		  } 
		    else (*vec)[(start_vec+k)*num_bits+i]='0';
		  }
	      }		
	  }

	/*inverse transformation*/
	if(inv_transform==1) //0:without inverse transformation, 1: with inverse transf 
	  {
	    MatrMult(1,H_size,H_size,coef,H_matrix,tmpVect);
	    for(k=H_size;k>=0;k--) 
	      {
		if(tmpVect[k]> 0) (*vec)[(start_vec+k)*num_bits+i]='1'; 
		else if(tmpVect[k]< 0) (*vec)[(start_vec+k)*num_bits+i]='0';
		else 
		  {
		    (*vec)[(start_vec+k)*num_bits+i]= (rand()+rand())%2+'0';//random 0 or 1
		    // printf("random selection ...........%d",(*vec)[m+row*8+k][i] );
		  }
		//printf("%3d",(*vec)[(start_vec+k)*num_bits+i]);
	      }
	    // printf("\n");
	  }
       
       /*flip each bit if necessary*/
       if(flip_bit == 1)
	 {
	   for(k=H_size;k>=0;k--) 
	     {
	       (*vec)[(start_vec+k)*num_bits+i] = '1'+'0'-(*vec)[(start_vec+k)*num_bits+i]; 
	     }
	 }
       
      }//for(row...   for(i...

  /*set the remainder vectors to random bits*/  
  for(i= num_bits*init_vec+num_bits*(j/H_size)*H_size-1;i<num_bits*(init_vec+new_vec); i++)
    (*vec)[i] = rand()%2+'0';
  
  return 1;
}

int random_expand(int *vec, int num_bits,int num_vec, int rdexp,char *vecfilename, int fix_hold
		  /*hold for fix time*/)
{
  int repeat, ttv;
  int i,j,k;
  FILE *vecfile;

  vecfile = fopen(vecfilename, "w");
  ttv = 0; 
  for(i=0;i<num_vec;i++)
    {
      repeat = rand()%rdexp;
      if(fix_hold == true) repeat = rdexp;
      ttv += repeat;
      for(k = 0; k<repeat; k++)
	{
	  for(j=0;j<num_bits;j++) fprintf(vecfile, "%c", vec[i*num_bits+j]);
	  fprintf(vecfile, "\n");
	}
    }
  fclose(vecfile);
  return ttv;
}

int random_expandV(int **from_array, int num_bits,int num_vec, int rdexp, int **to_array, int fix_hold
		   /*hold for fixed times*/)
{
  int repeat, ttv;
  int i,j,k;

  ttv = 0; 
  for(i=0;i<num_vec;i++)
    {
      repeat = rand()%rdexp;
      if(fix_hold == true) repeat = rdexp;
      for(k = 0; k<repeat; k++)
	{
	  for(j=0;j<num_bits;j++) 
	    {
	      (*to_array)[ttv*num_bits+j] = (*from_array)[i*num_bits+j];
	      if((*from_array)[i*num_bits+j]!='1' && (*from_array)[i*num_bits+j]!='0' )
		fprintf(resultsout, "%c!='0','1',ttv=%d,random_expandV\n", (*from_array)[i*num_bits+j],ttv+1);
	      if(sfsimdebug)
		fprintf(resultsout, "%d", (*from_array)[i*num_bits+j]);
	    }
	  if(sfsimdebug)
	    fprintf(resultsout, "\n");

	  ttv ++;
	  if(ttv == tmpvecsize)
	      resize_tmpvec(tmpvecsize, tmpvecsize*2, num_bits);
	}
    }
  return ttv;
}

  

