#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "all.h"

/***************************************************************************
  search if tgate is a fanout of sgate
****************************************************************************/
int searchfanout(sgate, tgate)
     NODE *sgate;
     NODE *tgate;
{ /* searchfanout */
  int i,j;

  j = nmbroutputs(sgate);

  for(i=0;i<j;++i)
    {
      if(fanoutgate(sgate, i) == tgate)
        return(i);
    }
  fprintf(resultsout, "searchfanout: Gate %d does not fan out to gate %d.\n",
	  gate_number(sgate), gate_number(tgate));
  fatal_error("searchfanout", FATAL_INTERNAL, NULL);
  exit (-1);

} /* searchfanout */

/***************************************************************************
  verify_ckt_topology() verifies the various connections of the ckt graph
          if they are proper or not
****************************************************************************/
void verify_ckt_topology (ckt)
     CIRCUIT *ckt;
{ /* verify_ckt_topology */  
  int i,j;
  int nmbr_inputs;
  int nmbr_outputs;
  int inp_nmbr;
  NODE *sgate, *tgate;
  NODE *gate;

  if(VERBOSE == TRUE) 
    fprintf(resultsout, "Verifying circuit topology ...\n");
  for(i=0;i<nmbr_pi(ckt);++i)
    {
      gate = PIs(ckt, i);
      nmbr_outputs = nmbroutputs(gate);
      if(nmbr_outputs < 0)
	{
#if defined(EDIF)
	    fprintf(resultsout,
	   "Illegal #of outputs -- pi index %d gate %s (%d) -- #outputs %d.\n",
		    i, compname (gate), gate_number(gate), nmbr_outputs);
#elif defined(L)
	    fprintf(resultsout,
	   "Illegal #of outputs -- pi index %d gate %s (%d) -- #outputs %d.\n",
		    i, compname (gate), gate_number(gate), nmbr_outputs);
#else
	    fprintf(resultsout,
	    "Illegal #of outputs -- pi index %d gate %d -- #outputs %d.\n",
		    i, gate_number(gate), nmbr_outputs);
#endif
	  fatal_error("verify_ckt_topology", GRAPH_VERIFY_ERROR, NULL);
	}
	  
      for(j=0;j<nmbr_outputs;++j)
	{
	  inp_nmbr = fanoutgateinpnmbr(gate,j);
	  tgate = fanoutgate(gate, j);
	  if(sourcegate(tgate, inp_nmbr) != gate)
	    {
#if defined(EDIF)
		fprintf(resultsout,
"Illegal fanout -- pi index %d gate# %s (%d) output %d, tgate gate# %s (%d).\n",
		      i, compname (gate), gate_number(gate), j,
			compname (tgate), gate_number(tgate));
#elif defined(L)
		fprintf(resultsout,
"Illegal fanout -- pi index %d gate# %s (%d) output %d, tgate gate# %s (%d).\n",
		      i, compname (gate), gate_number(gate), j,
			compname (tgate), gate_number(tgate));
#else
		fprintf(resultsout,
"Illegal fanout -- pi index %d gate# %s (%d) output %d, tgate gate# %s (%d).\n",
		      i, compname (gate), gate_number(gate), j,
			compname (tgate), gate_number(tgate));
#endif
	      fatal_error("verify_ckt_topology", GRAPH_VERIFY_ERROR, NULL);
	    }
	}
    }
  
  for(i=0;i<nmbr_gates(ckt);++i)
    {
      gate = gatesx(ckt, i);
      nmbr_outputs = nmbroutputs(gate);
      if(nmbr_outputs < 0)
	{
#if defined(EDIF)
	    fprintf(resultsout,
 "Illegal # of outputs -- gate index %d gate# %s (%d) -- #outputs %d.\n",
		  i, compname (gate), gate_number(gate), nmbr_outputs);
#elif defined(L)
	    fprintf(resultsout,
 "Illegal # of outputs -- gate index %d gate# %s (%d) -- #outputs %d.\n",
		  i, compname (gate), gate_number(gate), nmbr_outputs);
#else
	    fprintf(resultsout,
	 "Illegal # of outputs -- gate index %d gate# %d -- #outputs %d.\n",
		  i, gate_number(gate), nmbr_outputs);
#endif
	  fatal_error("verify_ckt_topology", GRAPH_VERIFY_ERROR, NULL);
	}
	  
      for(j=0;j<nmbr_outputs;++j)
	{
	  inp_nmbr = fanoutgateinpnmbr(gate,j);
	  tgate = fanoutgate(gate, j);
	  if(sourcegate(tgate, inp_nmbr) != gate)
	    {
#if defined(EDIF)
		fprintf(resultsout,
 "Illegal fanout -- gate index %d gate# %s (%d) output %d, tgate # %s (%d).\n",
			i, compname (gate), gate_number(gate), j,
			compname (tgate), gate_number(tgate));
#elif defined(L)
		fprintf(resultsout,
 "Illegal fanout -- gate index %d gate# %s (%d) output %d, tgate # %s (%d).\n",
			i, compname (gate), gate_number(gate), j,
			compname (tgate), gate_number(tgate));
#else
		fprintf(resultsout,
      "Illegal fanout -- gate index %d gate# %d output %d, tgate # %d.\n",
		      i, gate_number(gate), j, gate_number(tgate));
#endif
	      fatal_error("verify_ckt_topology", GRAPH_VERIFY_ERROR, NULL);
	    }
	}

      nmbr_inputs = nmbrinputs(gate);
      if(nmbr_inputs <= 0)
	{
#if defined(EDIF)
	    fprintf(resultsout,
  "Illegal # of inputs -- gate index %d gate# %s (%d) -- #inputs %d.\n",
		  i, compname (gate), gate_number(gate), nmbr_inputs);
#elif defined(L)
	    fprintf(resultsout,
  "Illegal # of inputs -- gate index %d gate# %s (%d) -- #inputs %d.\n",
		  i, compname (gate), gate_number(gate), nmbr_inputs);
#else
	    fprintf(resultsout,
	  "Illegal # of inputs -- gate index %d gate# %d--#inputs %d.\n",
		  i, gate_number(gate), nmbr_inputs);
#endif
	      fatal_error("verify_ckt_topology", GRAPH_VERIFY_ERROR, NULL);
	}
      for(j=0;j<nmbr_inputs;++j)
	{
	  sgate = sourcegate(gate,j);
	  if(searchfanout(sgate, gate) == -1)
	    {
#if defined(EDIF)
		fprintf(resultsout,
       "Illegal source gate -- gate index %d gate# %s (%d) source gate %d.\n",
		      i, compname (gate), gate_number (gate), j);
#elif defined(L)
		fprintf(resultsout,
       "Illegal source gate -- gate index %d gate# %s (%d) source gate %d.\n",
		      i, compname (gate), gate_number (gate), j);
#else
		fprintf(resultsout,
       "Illegal source gate -- gate index %d gate# %d source gate %d.\n",
		      i, gate_number (gate), j);
#endif
	      fatal_error("verify_ckt_topology", GRAPH_VERIFY_ERROR, NULL);
	    }
	}
    }

  for(i=0;i<nmbr_po(ckt);++i)
    {
      gate = POs(ckt, i);
      nmbr_inputs = nmbrinputs(gate);
      if(nmbr_inputs != 1)
	{
#if defined(EDIF)
	    fprintf(resultsout,
	  "Illegal # of inputs -- po index %d gate# %s (%d) -- #inputs %d.\n",
		  i, compname (gate), gate_number (gate), nmbr_inputs);
#elif defined(L)
	    fprintf(resultsout,
	  "Illegal # of inputs -- po index %d gate# %s (%d) -- #inputs %d.\n",
		  i, compname (gate), gate_number (gate), nmbr_inputs);
#else
	    fprintf(resultsout,
	  "Illegal # of inputs -- po index %d gate# %d -- #inputs %d.\n",
		  i, gate_number (gate), nmbr_inputs);
#endif
	      fatal_error("verify_ckt_topology", GRAPH_VERIFY_ERROR, NULL);
	}
      for(j=0;j<nmbr_inputs;++j)
	{
	  sgate = sourcegate(gate,j);
	  if(searchfanout(sgate, gate) == -1)
	    {
		fprintf(resultsout,
			"Illegal source gate--po index %d source gate %d.\n",
		      i, j);
	      fatal_error("verify_ckt_topology", GRAPH_VERIFY_ERROR, NULL);
	    }
	}
    }
  if(VERBOSE == TRUE)
    fprintf(resultsout, "done\n");

} /* verify_ckt_topology */
